{%- extends "basic/layout.html" %}
{%- block doctype %}
    <!doctype html>
{% endblock %}
{%- block extrahead %}
    <link rel="icon" type="image/png" href="{{ pathto('_static/errbot.png', 1) }}"/>
    <link rel="apple-touch-icon" href="{{ pathto('_static/errbot.png', 1) }}"/>
    <link rel="stylesheet" href="{{ pathto('_static/' ~ "fancybox/jquery.fancybox.css", 1) }}" type="text/css"
          media="screen"/>
    <script type="text/javascript" src="{{ pathto('_static/' ~ "fancybox/jquery.fancybox.pack.js", 1) }}"></script>
    <meta name="viewport" content="width=device-width, initial-scale=0.9, maximum-scale=0.9"/>
    <meta name="description" content="A Python chatbot with a rich and friendly API and an emphasis on developer productivity">
    <meta name="keywords" content="err,errbot,chat,chatbot,bot,python,devops,chatops,automation"/>
    <meta name="generator" content="Sphinx">
    <script type="text/javascript">
        // Google Analytics
        var _gaq = _gaq || [];
        _gaq.push(['_setAccount', 'UA-32534172-1']);
        _gaq.push(['_trackPageview']);

        (function () {
            var ga = document.createElement('script');
            ga.type = 'text/javascript';
            ga.async = true;
            ga.src = ('https:' == document.location.protocol ? 'https://ssl' : 'http://www') + '.google-analytics.com/ga.js';
            var s = document.getElementsByTagName('script')[0];
            s.parentNode.insertBefore(ga, s);
        })();

        // Gitter button
        ((window.gitter = {}).chat = {}).options = {
             room: 'errbotio/errbot'
        };

        // Piwik
        var _paq = _paq || [];
        _paq.push(["setDomains", ["*.errbot.io","errbot.readthedocs.io"]]);
        _paq.push(['trackPageView']);
        _paq.push(['enableLinkTracking']);
        (function() {
            var u="//www.nickgroenen.nl/";
            _paq.push(['setTrackerUrl', u+'pa.php']);
            _paq.push(['setSiteId', '2']);
            var d=document, g=d.createElement('script'), s=d.getElementsByTagName('script')[0];
            g.type='text/javascript'; g.async=true; g.defer=true; g.src=u+'pa.js'; s.parentNode.insertBefore(g,s);
        })();
   </script>
   <noscript><p><img src="//www.nickgroenen.nl/pa.php?idsite=2&rec=1" style="border:0;" alt=""/></p></noscript>
   <script src="https://sidecar.gitter.im/dist/sidecar.v1.js" async defer></script>
{% endblock %}

{%- block sidebarlogo %}
    <a class="logo" href="{{ pathto(master_doc) }}" title="Back to home page">
        <img class="logo" src="{{ pathto('_static/errbot.png', 1) }}" width="100" height="100" alt="Logo of Errbot"/>
    </a>
    <div class="ghbuttons">
        <iframe src="//ghbtns.com/github-btn.html?user=errbotio&repo=errbot&type=watch&count=true"
                allowtransparency="true" frameborder="0" scrolling="0" width="95" height="20"></iframe>
        <iframe src="//ghbtns.com/github-btn.html?user=errbotio&repo=errbot&type=fork&count=true" allowtransparency="true"
                frameborder="0" scrolling="0" width="95" height="20"></iframe>
    </div>
{%- endblock %}

{%- block relbar1 %}{{ super() }}{% endblock %}
{%- block relbar2 %}
    <div class="relbar2">
        {{ super() }}
    </div>
{%- endblock %}

{% block header %}
    {{ super() }}
    {% if pagename == 'index' %}
        <div class=indexwrapper>
    {% endif %}
{% endblock %}

{%- block footer %}
    <div class="footer">
        &copy; Copyright {{ copyright }}.
        {%- if last_updated %}
          Last updated on {{ last_updated }}.
        {%- endif %}
        Created using <a href="http://sphinx.pocoo.org/">Sphinx</a>.
    </div>
    {% if pagename == 'index' %}
        </div>
    {% endif %}
{%- endblock %}

{%- block content %}
    {{ super() }}
    <script>
        $(document).ready(function () {
            $('.fancybox').fancybox({
                padding: 0,
                openEffect: 'elastic'
            });
        });
    </script>
{%- endblock %}
