<h3>Related Topics</h3>
<ul>
    <li><a href="{{ pathto(master_doc) }}">Homepage</a></li>
    <li><a href="{{ pathto("genindex") }}">Index</a></li>
    <li><a href="{{ pathto("py-modindex") }}">Module index</a></li>
    <li><a href="{{ pathto("modules") }}">API reference</a></li>
    {%- for parent in parents %}
        <li>
        <a href="{{ parent.link|e }}">{{ parent.title }}</a>
        <ul>
    {%- endfor %}
    {%- if prev %}
        <li>Previous: <a href="{{ prev.link|e }}" title="{{ _('previous chapter') }}">{{ prev.title }}</a></li>
    {%- endif %}
    {%- if next %}
        <li>Next: <a href="{{ next.link|e }}" title="{{ _('next chapter') }}">{{ next.title }}</a></li>
    {%- endif %}
    {%- for parent in parents %}
        </ul>
        </li>
    {%- endfor %}
</ul>
