body {
  background-repeat: no-repeat;
  background-position: center center;
  background-attachment:fixed;
  background-size: contain;
  background-color: #FFFFFF;
  margin:0;
}

.receiving {
  background-color: #f8ebdf;
  border-color: #ed9d53;
  margin:5px;
  padding-top:5px;
  padding-left:10px;
  padding-right:10px;
  padding-bottom:10px;
  border-style:solid;
  border-width: 1px;
  border-bottom-left-radius:5px;
  border-bottom-right-radius:5px;
  border-top-right-radius:5px;
  border-top-left-radius:5px;
  opacity:0.8;
}

.sending {
  background-color: #dfdef7;
  border-color: #8fa9e6;
  margin:5px;
  padding-top:5px;
  padding-left:10px;
  padding-right:10px;
  padding-bottom:10px;
  border-style:solid;
  border-width: 1px;
  border-bottom-left-radius:5px;
  border-bottom-right-radius:5px;
  border-top-right-radius:5px;
  border-top-left-radius:5px;
  font-weight:bold;
  opacity:0.8;
}

pre {
  white-space: pre-line; 
  font-size:10px;
  border-width:1px;
  background-color:#DDDDDD;
  border-color: #555555;
  border-bottom-left-radius:5px;
  border-bottom-right-radius:5px;
  border-top-right-radius:5px;
  border-top-left-radius:5px;
 }

table {
  margin:10px;
  border-bottom-left-radius:5px;
  border-bottom-right-radius:5px;
  border-top-right-radius:5px;
  border-top-left-radius:5px;
  border-collapse: collapse;
  width: 80%;
}

td{
  vertical-align:middle;
  font-size:14px;
  text-align:center;
  margin-left: 30px;
  font-family:Arial, sans-serif;
  color:#333333;
}

tr th{
  background:#9667b2;
  text-align:center;
  font-size:16px;
  font-family:Arial, sans-serif;
  font-weight:bold;
  color:#FFFFFF;
}

tr:last-child td:last-child {
  border-bottom-right-radius:5px;
}

table tr:first-child th:first-child {
  border-top-left-radius:5px;
}

table tr:first-child th:last-child {
  border-top-right-radius:5px;
}

tr:last-child td:first-child{
  border-bottom-left-radius:5px;
}

tr:nth-child(odd) td {
  background-color:#FFF3DC;
}

tr:nth-child(even) td {
  background-color:#FFF7EC;
}

*[color="red"] {
    color: red;
}

*[color="green"] {
    color: green;
}

*[color="yellow"] {
    color: orange;
}

*[color="blue"] {
    color: blue;
}

*[color="white"] {
    color: #FFFFFF;
}

*[bgcolor="red"] {
    background-color: lightsalmon;
}

*[bgcolor="green"] {
    background-color: lightgreen;
}

*[bgcolor="yellow"] {
    background-color: lightgoldenrodyellow;
}

*[bgcolor="cyan"] {
    background-color: aqua;
}

*[bgcolor="blue"] {
    background-color: lightblue;
}

*[bgcolor="white"] {
    background-color: azure;
}

*[bgcolor="black"] {
    background-color: black;
}

h1 { color: #111; font-family: 'Helvetica Neue', sans-serif; font-size: 50px; font-weight: bold; letter-spacing: -1px; line-height: 1; text-align: center; }

h2 { color: #111; font-family: 'Open Sans', sans-serif; font-size: 30px; font-weight: 300; line-height: 32px; margin: 0 0 20px; text-align: center; }

h3 { color: #111; font-family: 'Arial Unicode MS', sans-serif; font-size: 25px; font-weight: 200; line-height: 27px; margin: 0 0 10px; text-align: center; }

body { color: #111; font-family: 'Arial Unicode MS', sans-serif; font-size: 14px; line-height: 24px; margin: 0 0 24px;  }
